#define WINVER 0x0500
#include <windows.h>
#include <math.h>



void GetRGBPalette(HBITMAP hBitmap, RGBQUAD *ColorTable)
{
	HDC  hdcMem;

	hdcMem = CreateCompatibleDC(NULL);
	SelectObject(hdcMem, hBitmap);
  GetDIBColorTable(hdcMem, 0, 256, ColorTable);  // Aktuelle Farbtabelle bernehmen
	DeleteDC(hdcMem);
}


/********************************************************************************/
/*                                                                              */
/*  Routine zum Anlegen einer neuen Graustufen-Bitmap im Speicher (DibSection)  */
/*                                                                              */
/********************************************************************************/

HBITMAP NewDIB(long Width, long Height, RGBQUAD *RGBPalette)
{
  HBITMAP          hBitmap;
	BITMAPINFO      *pbmi;
	BYTE            *pBits;
	long             Counter;

	// BMI-Struktur mit Daten fllen
	pbmi = (tagBITMAPINFO*) malloc (sizeof (BITMAPINFOHEADER)+256*sizeof(RGBQUAD));
	if (!pbmi)    // Nicht genug Speicher???
	{
		return NULL;
	}
	pbmi->bmiHeader.biSize          = sizeof (BITMAPINFOHEADER);
	pbmi->bmiHeader.biWidth         = Width;  // bernimm Gre des Bitmaps
	pbmi->bmiHeader.biHeight        = Height; // aus der zu ffnenden Datei
	pbmi->bmiHeader.biPlanes        = 1;  // Immer 1
  pbmi->bmiHeader.biBitCount      = 8;  // 256 Graustufen
	pbmi->bmiHeader.biCompression   = 0;  // unkomprimiert
	pbmi->bmiHeader.biClrImportant  = 0;  // Alle Farben wichtig
	pbmi->bmiHeader.biClrUsed       = 256;  // Alle 256 Farben werden benutzt
	pbmi->bmiHeader.biSizeImage     = 4*((Width*8+31)/32)*Height;  // Gre der Pixeldaten in Bytes
	pbmi->bmiHeader.biXPelsPerMeter = 0;  // Mastab Pixel / Meter
	pbmi->bmiHeader.biYPelsPerMeter = 0;  // Mastab Pixel / Meter
	if (RGBPalette)
	{
		for (Counter=0; Counter<=255; Counter++)
		{
			pbmi->bmiColors[Counter].rgbBlue = RGBPalette[Counter].rgbBlue;
			pbmi->bmiColors[Counter].rgbGreen = RGBPalette[Counter].rgbGreen;
			pbmi->bmiColors[Counter].rgbRed = RGBPalette[Counter].rgbRed;
			pbmi->bmiColors[Counter].rgbReserved = 0;
		}
	}
	else
	{
		for (Counter=0; Counter<=255; Counter++)
		{
			pbmi->bmiColors[Counter].rgbBlue = (BYTE) Counter;
			pbmi->bmiColors[Counter].rgbGreen = (BYTE) Counter;
			pbmi->bmiColors[Counter].rgbRed = (BYTE) Counter;
			pbmi->bmiColors[Counter].rgbReserved = 0;
		}
	}
	// DIB Section auf Basis dieser Daten anlegen
  hBitmap = CreateDIBSection (NULL, pbmi, DIB_RGB_COLORS, (void**) &pBits, NULL, 0) ;
  free (pbmi);  // Nur Speicher der BMI-Struktur wieder freigeben
  return hBitmap;  // Wenn nicht erfolgreich ist hBitmap==NULL
}





/********************************************************************/
/*                                                                  */
/*  Routine zum Einladen einer Bitmap in den Speicher (DibSection)  */
/*                                                                  */
/********************************************************************/

HBITMAP CreateDibSectionFromDibFile (PTSTR szFileName, BOOL &IsGrayDIB)
{
     BITMAPFILEHEADER bmfh;
     BOOL             bSuccess;
     DWORD            dwInfoSize, dwBytesRead, dwClrUsed;
     HANDLE           hFile;
     HBITMAP          hBitmap;
		 BITMAPINFO      *pbmi, *pFILEbmi;
		 BYTE            *pBits,*pFILEBits;
		 long             Counter;

     // Datei fr Lesezugriffe ffnen (und gegen Schreibaktionen anderer Prozesse sperren)
     hFile = CreateFile (szFileName, GENERIC_READ, FILE_SHARE_READ,
                         NULL, OPEN_EXISTING, 0, NULL) ;

     if (hFile == INVALID_HANDLE_VALUE)
          return NULL ;

     // BITMAPFILEHEADER einlesen
     bSuccess = ReadFile (hFile, &bmfh, sizeof (BITMAPFILEHEADER), 
                          &dwBytesRead, NULL) ;

     if (!bSuccess || (dwBytesRead != sizeof (BITMAPFILEHEADER))         
                   || (bmfh.bfType != * (WORD *) "BM"))
     {
          CloseHandle (hFile) ;
          return NULL ;
     }

     // Platz fr eine BITMAPINFO-Struktur belegen und Daten einlesen
     dwInfoSize = bmfh.bfOffBits - sizeof (BITMAPFILEHEADER) ;

		 pFILEbmi = (tagBITMAPINFO*) malloc (dwInfoSize) ;
		 pbmi     = (tagBITMAPINFO*) malloc (sizeof (BITMAPINFOHEADER)+256*sizeof(RGBQUAD));

     bSuccess = ReadFile (hFile, pFILEbmi, dwInfoSize, &dwBytesRead, NULL) ;

     if (dwBytesRead != dwInfoSize) bSuccess = FALSE;
		 if (pFILEbmi->bmiHeader.biPlanes != 1) bSuccess = FALSE;
		 if (pFILEbmi->bmiHeader.biBitCount != 8) bSuccess = FALSE;
		 if (pFILEbmi->bmiHeader.biCompression != 0) bSuccess = FALSE;

		 if (!bSuccess)
     {
          free (pFILEbmi);
			    free (pbmi);
          CloseHandle (hFile);
          return NULL;
     }

		 pbmi->bmiHeader.biSize          = sizeof (BITMAPINFOHEADER);
		 pbmi->bmiHeader.biWidth         = pFILEbmi->bmiHeader.biWidth;  // bernimm Gre des Bitmaps
		 pbmi->bmiHeader.biHeight        = abs(pFILEbmi->bmiHeader.biHeight); // aus der zu ffnenden Datei
		 pbmi->bmiHeader.biPlanes        = 1;  // Immer 1
     pbmi->bmiHeader.biBitCount      = 8;  // 256 Graustufen
		 pbmi->bmiHeader.biCompression   = 0;  // unkomprimiert
		 pbmi->bmiHeader.biClrImportant  = 0;  // Alle Farben wichtig
		 pbmi->bmiHeader.biClrUsed       = 256;  // Alle 256 Farben werden benutzt
		 pbmi->bmiHeader.biSizeImage     = 4*((pbmi->bmiHeader.biWidth*8+31)/32)*pbmi->bmiHeader.biHeight;  // Gre der Pixeldaten in Bytes
		 pbmi->bmiHeader.biXPelsPerMeter = pFILEbmi->bmiHeader.biXPelsPerMeter;  // Mastab Pixel / Meter
		 pbmi->bmiHeader.biYPelsPerMeter = pFILEbmi->bmiHeader.biYPelsPerMeter;  // Mastab Pixel / Meter

	   IsGrayDIB = TRUE;
		 for (Counter=0; Counter<=255; Counter++)
		 {
			 if ( (pbmi->bmiColors[Counter].rgbBlue = pFILEbmi->bmiColors[Counter].rgbBlue) != (BYTE) Counter)    IsGrayDIB = FALSE;
			 if ( (pbmi->bmiColors[Counter].rgbGreen = pFILEbmi->bmiColors[Counter].rgbGreen) != (BYTE) Counter)  IsGrayDIB = FALSE;
			 if ( (pbmi->bmiColors[Counter].rgbRed = pFILEbmi->bmiColors[Counter].rgbRed) != (BYTE) Counter)      IsGrayDIB = FALSE;
			 pbmi->bmiColors[Counter].rgbReserved = 0;
		 }

		 // DIB Section auf Basis dieser Daten anlegen
     hBitmap = CreateDIBSection (NULL, pbmi, DIB_RGB_COLORS, (void**) &pBits, NULL, 0) ;

     if (hBitmap == NULL)
     {
          free (pbmi) ;
          free (pFILEbmi);
					CloseHandle (hFile) ;
          return NULL ;
     }

     // DIB-Pixeldaten in den von CreateDIBSection belegten Speicherblock einlesen
     pFILEBits = (BYTE*) malloc (bmfh.bfSize - bmfh.bfOffBits);
		 ReadFile (hFile, pFILEBits, bmfh.bfSize - bmfh.bfOffBits, &dwBytesRead, NULL) ;

		 dwClrUsed = pFILEbmi->bmiHeader.biClrUsed;
		 if (!dwClrUsed)  dwClrUsed = 256;

		 CopyMemory(pBits, pFILEBits, pbmi->bmiHeader.biSizeImage);
		 free (pFILEBits);
		 free (pbmi);
		 free (pFILEbmi);
     CloseHandle (hFile) ;

     return hBitmap ;
}




/***************************************************************/
/*                                                             */
/*  Routine zum Speichern der aktuellen Bitmap in einer Datei  */
/*                                                             */
/***************************************************************/

BOOL WriteDibSectionToDibFile (HBITMAP hBitmap, TCHAR *szFileName)
{
  DIBSECTION       dibsection;
	BITMAPFILEHEADER bmfh;
	BITMAPINFOHEADER *pbmih;
//	HDC              hdcMem;
	RGBQUAD          ActPal[256];
	BOOL             bSuccess;
  HANDLE           hFile;
	DWORD						 dwTotalSize, dwBytesWritten;
	BYTE             *pTmpMem;
	long			       Counter;

	if (!GetObject(hBitmap, sizeof (DIBSECTION), &dibsection))  return FALSE;
	pbmih = &dibsection.dsBmih;
	hFile = CreateFile (szFileName, GENERIC_WRITE, 0, NULL,
		                  CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);

	if (hFile == INVALID_HANDLE_VALUE)  return FALSE;

	dwTotalSize = sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFOHEADER) + 256*sizeof(RGBQUAD) + 4*((pbmih->biWidth*8+31)/32)*pbmih->biHeight;
	
	bmfh.bfType      = * (WORD *) "BM";
	bmfh.bfSize      = dwTotalSize;
	bmfh.bfReserved1 = 0;
	bmfh.bfReserved2 = 0;
	bmfh.bfOffBits   = bmfh.bfSize - 4*((pbmih->biWidth*8+31)/32)*pbmih->biHeight;

	bSuccess = WriteFile (hFile, &bmfh, sizeof (BITMAPFILEHEADER), &dwBytesWritten, NULL);
	if (!bSuccess || (dwBytesWritten != sizeof (BITMAPFILEHEADER)))
	{
		CloseHandle (hFile);
		DeleteFile (szFileName);
		return FALSE;
	}

	bSuccess = WriteFile (hFile, pbmih, sizeof (BITMAPINFOHEADER), &dwBytesWritten, NULL);
	if (!bSuccess || (dwBytesWritten != sizeof (BITMAPINFOHEADER)))
	{
		CloseHandle (hFile);
		DeleteFile (szFileName);
		return FALSE;
	}

	pTmpMem = (BYTE*) malloc (4*256);
	if (!pTmpMem)
	{
		CloseHandle (hFile);
		DeleteFile (szFileName);
		return FALSE;
	}
  GetRGBPalette(hBitmap, ActPal);
	for (Counter=0; Counter<256; Counter++)
	{
	  pTmpMem[Counter*4] = ActPal[Counter].rgbBlue;
	  pTmpMem[Counter*4+1] = ActPal[Counter].rgbGreen;
	  pTmpMem[Counter*4+2] = ActPal[Counter].rgbRed;
	  pTmpMem[Counter*4+3] = 0;
	}
	bSuccess = WriteFile (hFile, pTmpMem, 4*256, &dwBytesWritten, NULL);
  free(pTmpMem);
	if (!bSuccess || (dwBytesWritten != 4*256))
	{
		CloseHandle (hFile);
		DeleteFile (szFileName);
		return FALSE;
	}

	bSuccess = WriteFile (hFile, dibsection.dsBm.bmBits, 4*((pbmih->biWidth*8+31)/32)*pbmih->biHeight, &dwBytesWritten, NULL);
	if (!bSuccess || (dwBytesWritten != 4*((pbmih->biWidth*8+31)/32)*pbmih->biHeight))
	{
		CloseHandle (hFile);
		DeleteFile (szFileName);
		return FALSE;
	}

	CloseHandle (hFile);
  return TRUE;
}




/************************************************************************/
/*                                                                      */
/*  Routine zum Ermitteln der Gre einer Bitmap in der Zwischenablage  */
/*                                                                      */
/************************************************************************/

void GetDIBSizeFromClipboard(HWND hwnd, long &Width, long &Height)
{
	HGLOBAL			hGlobal;
	BYTE				*pGlobal;
	BITMAPINFO  *pCLIPbmi;

  Width = Height = 0;   // Bleiben 0, wenn Fehler auftritt
	// Zwischenablage ffnen
	OpenClipboard(hwnd);
	hGlobal = GetClipboardData(CF_DIB);
	if (!hGlobal)
	{
		CloseClipboard();
		return;
	}
	pGlobal = (BYTE *) GlobalLock(hGlobal);
	if (!pGlobal)
	{
		GlobalUnlock(hGlobal);
		CloseClipboard();
		return;
	}
	pCLIPbmi = (BITMAPINFO *) pGlobal;

  // Gre aus der Bitmap-Header-Struktur in Width und Height kopieren, wenn Bitmap ok
	if (pCLIPbmi->bmiHeader.biBitCount==8  &&  pCLIPbmi->bmiHeader.biCompression==0  &&  pCLIPbmi->bmiHeader.biPlanes==1)
	{
		Width  = pCLIPbmi->bmiHeader.biWidth;
		Height = pCLIPbmi->bmiHeader.biHeight;
	}

	// Zwischenablage schlieen
	GlobalUnlock (hGlobal);
	CloseClipboard();
	return;
}





/**************************************************************/
/*                                                            */
/*  Routine zum Einfgen einer Bitmap aus der Zwischenablage  */
/*                                                            */
/**************************************************************/

HBITMAP PasteDIBfromClipboard(HWND hwnd, BOOL &IsGrayDIB)
{

	HGLOBAL			hGlobal;
	BYTE				* pGlobal;
  BOOL        bSuccess = TRUE;
  DWORD       dwInfoSize, dwClrUsed;
  HBITMAP     hBitmap;
	BITMAPINFO  *pbmi, *pCLIPbmi;
	BYTE        *pBits,*pCLIPBits;
	long        Counter;

  // Zwischenablage ffnen
	OpenClipboard (hwnd);

	hGlobal = GetClipboardData (CF_DIB);
	if (!hGlobal)
	{
		CloseClipboard();
		return NULL;
	}
	pGlobal = (BYTE *) GlobalLock (hGlobal);

	if (!pGlobal)
	{
		GlobalUnlock(hGlobal);
		CloseClipboard();
		return NULL;
	}

	pCLIPbmi = (BITMAPINFO *) pGlobal;

  // Platz fr eine BITMAPINFO-Struktur belegen und Daten einlesen
  dwInfoSize = pCLIPbmi->bmiHeader.biSize;

	if (dwInfoSize != sizeof (BITMAPCOREHEADER) &&
		  dwInfoSize != sizeof (BITMAPINFOHEADER) &&
			dwInfoSize != sizeof (BITMAPV4HEADER) &&
			dwInfoSize != sizeof (BITMAPV5HEADER) )      bSuccess = FALSE;

	if (pCLIPbmi->bmiHeader.biCompression != 0 ||
		  pCLIPbmi->bmiHeader.biBitCount != 8 ||
			pCLIPbmi->bmiHeader.biPlanes != 1)           bSuccess = FALSE;

	if (!bSuccess)
	{
		GlobalUnlock (hGlobal);
		CloseClipboard();
		return NULL;
	}
	
	pbmi     = (tagBITMAPINFO*) malloc (sizeof (BITMAPINFOHEADER)+256*sizeof(RGBQUAD));

	pbmi->bmiHeader.biSize          = sizeof (BITMAPINFOHEADER);
	pbmi->bmiHeader.biWidth         = pCLIPbmi->bmiHeader.biWidth;  // bernimm Gre des Bitmaps
	pbmi->bmiHeader.biHeight        = abs(pCLIPbmi->bmiHeader.biHeight); // aus der zu ffnenden Datei
	pbmi->bmiHeader.biPlanes        = 1;  // Immer 1
  pbmi->bmiHeader.biBitCount      = 8;  // 256 Graustufen
	pbmi->bmiHeader.biCompression   = 0;  // unkomprimiert
	pbmi->bmiHeader.biClrImportant  = 0;  // Alle Farben wichtig
	pbmi->bmiHeader.biClrUsed       = 256;  // Alle 256 Farben werden benutzt
	pbmi->bmiHeader.biSizeImage     = 4*((pbmi->bmiHeader.biWidth*8+31)/32)*pbmi->bmiHeader.biHeight;  // Gre der Pixeldaten in Bytes
	pbmi->bmiHeader.biXPelsPerMeter = pCLIPbmi->bmiHeader.biXPelsPerMeter;  // Mastab Pixel / Meter
	pbmi->bmiHeader.biYPelsPerMeter = pCLIPbmi->bmiHeader.biYPelsPerMeter;  // Mastab Pixel / Meter

  IsGrayDIB = TRUE;
  for (Counter=0; Counter<=255; Counter++)
	{
		if ( (pbmi->bmiColors[Counter].rgbBlue = pCLIPbmi->bmiColors[Counter].rgbBlue) != (BYTE) Counter)    IsGrayDIB = FALSE;
		if ( (pbmi->bmiColors[Counter].rgbGreen = pCLIPbmi->bmiColors[Counter].rgbGreen) != (BYTE) Counter)  IsGrayDIB = FALSE;
		if ( (pbmi->bmiColors[Counter].rgbRed = pCLIPbmi->bmiColors[Counter].rgbRed) != (BYTE) Counter)      IsGrayDIB = FALSE;
		pbmi->bmiColors[Counter].rgbReserved = 0;
	}

	// DIB Section auf Basis dieser Daten anlegen
  hBitmap = CreateDIBSection (NULL, pbmi, DIB_RGB_COLORS, (void**) &pBits, NULL, 0) ;

  if (hBitmap == NULL)
  {
    free (pbmi) ;
		GlobalUnlock (hGlobal);
		CloseClipboard();
    return NULL ;
  }

  // DIB-Pixeldaten in den von CreateDIBSection belegten Speicherblock einlesen
  dwClrUsed = pCLIPbmi->bmiHeader.biClrUsed;
	if (!dwClrUsed)  dwClrUsed = 256;
	pCLIPBits = (BYTE *) pCLIPbmi + dwInfoSize+4*dwClrUsed;

  CopyMemory(pBits, pCLIPBits, pbmi->bmiHeader.biSizeImage);
	free (pbmi);
	GlobalUnlock (hGlobal);
	CloseClipboard();

  return hBitmap ;
}




/*********************************************************************/
/*                                                                   */
/*  Routine zum Kopieren der aktuellen Bitmap in die Zwischenablage  */
/*                                                                   */
/*********************************************************************/

BOOL CopyDIBtoClipboard(HWND hwnd, HBITMAP hBitmap, RECT SelRect)
{
  DIBSECTION dibsection;
	RGBQUAD    ActPal[256];
	HGLOBAL    hGlobal;
	BYTE     * pGlobal;
	long			 x, y, WidthBytes;

	GetObject(hBitmap, sizeof (DIBSECTION), &dibsection);
	WidthBytes = dibsection.dsBm.bmWidthBytes;
	dibsection.dsBm.bmWidth = dibsection.dsBmih.biWidth = SelRect.right-SelRect.left+1;
	dibsection.dsBm.bmWidthBytes = 4*((dibsection.dsBmih.biWidth*8+31)/32);
	dibsection.dsBm.bmHeight = dibsection.dsBmih.biHeight = SelRect.top-SelRect.bottom+1;
	dibsection.dsBmih.biSizeImage = dibsection.dsBmih.biSize+dibsection.dsBm.bmWidthBytes*dibsection.dsBm.bmHeight+256*4;

	hGlobal = GlobalAlloc (GHND | GMEM_SHARE, dibsection.dsBmih.biSize+dibsection.dsBm.bmWidthBytes*dibsection.dsBm.bmHeight+256*4);
	pGlobal = (BYTE*) GlobalLock (hGlobal);
	CopyMemory(pGlobal, (BYTE*) &dibsection.dsBmih, dibsection.dsBmih.biSize);
	GetRGBPalette(hBitmap, ActPal);
	for (x=0; x<256; x++)    // Farbtabellen mit Graustufen initialisieren
	{
	  pGlobal[dibsection.dsBmih.biSize+x*4]   = ActPal[x].rgbBlue;
	  pGlobal[dibsection.dsBmih.biSize+x*4+1] = ActPal[x].rgbGreen;
	  pGlobal[dibsection.dsBmih.biSize+x*4+2] = ActPal[x].rgbRed;
	  pGlobal[dibsection.dsBmih.biSize+x*4+3] = 0;
	}   
	for (y=SelRect.bottom; y<=SelRect.top; y++)
	{
		for (x=SelRect.left; x<=SelRect.right; x++)
		{
			pGlobal[1024+dibsection.dsBmih.biSize+(y-SelRect.bottom)*dibsection.dsBm.bmWidthBytes+(x-SelRect.left)] = ((BYTE*) dibsection.dsBm.bmBits)[y*WidthBytes+x];
		}
		while (x-SelRect.left<dibsection.dsBm.bmWidthBytes) pGlobal[1024+dibsection.dsBmih.biSize+(y-SelRect.bottom)*dibsection.dsBm.bmWidthBytes+(x++-SelRect.left)]=0;
	}
	GlobalUnlock(hGlobal);
  OpenClipboard(hwnd);
	EmptyClipboard();
	SetClipboardData(CF_DIB, hGlobal);
	CloseClipboard();
  return TRUE;
}



void SetStandardPalette(HBITMAP hBitmap)
{
	RGBQUAD StdPal[256];
	int     counter;
	HDC     hdcMem;

	for (counter = 0; counter < 256; counter++)
	{
		StdPal[counter].rgbRed = counter;
		StdPal[counter].rgbGreen = counter;
		StdPal[counter].rgbBlue = counter;
		StdPal[counter].rgbReserved = 0;
	}
	hdcMem = CreateCompatibleDC(NULL);
	SelectObject(hdcMem, hBitmap);
  SetDIBColorTable(hdcMem, 0, 256, StdPal);
	DeleteDC(hdcMem);
}


void SetGrayPalette(HBITMAP hBitmap, BYTE *GrayValues)
{
	RGBQUAD StdPal[256];
	int     counter;
	HDC     hdcMem;

	for (counter = 0; counter < 256; counter++)
	{
		StdPal[counter].rgbRed = GrayValues[counter];
		StdPal[counter].rgbGreen = GrayValues[counter];
		StdPal[counter].rgbBlue = GrayValues[counter];
		StdPal[counter].rgbReserved = 0;
	}
	hdcMem = CreateCompatibleDC(NULL);
	SelectObject(hdcMem, hBitmap);
  SetDIBColorTable(hdcMem, 0, 256, StdPal);
	DeleteDC(hdcMem);
}


void SetRGBPalette(HBITMAP hBitmap, RGBQUAD *ColorTable)
{
	HDC     hdcMem;

	hdcMem = CreateCompatibleDC(NULL);
	SelectObject(hdcMem, hBitmap);
  SetDIBColorTable(hdcMem, 0, 256, ColorTable);
	DeleteDC(hdcMem);
}


void SetSingleRGBPaletteEntry(HBITMAP hBitmap, BYTE EntryNr, unsigned long rgbColor)
{
	HDC     hdcMem;
	RGBQUAD rgbQuad;

	rgbQuad.rgbReserved = 0;
	rgbQuad.rgbRed = BYTE(rgbColor & 0xFF);
	rgbQuad.rgbGreen = BYTE((rgbColor & 0xFF00) >> 8);
	rgbQuad.rgbBlue = BYTE((rgbColor & 0xFF0000) >> 16);
	hdcMem = CreateCompatibleDC(NULL);
	SelectObject(hdcMem, hBitmap);
  SetDIBColorTable(hdcMem, EntryNr, 1, &rgbQuad);
	DeleteDC(hdcMem);
}


BOOL LoadRGBPalette(char *szFileName, RGBQUAD *LoadPal)
{
	HANDLE  hFile;
	char    *szBuf;
	DWORD   dwLength, Offset;
	int     i, iTmp;

	hFile = CreateFile(szFileName, GENERIC_READ, 0, NULL, OPEN_EXISTING,
		                 FILE_SHARE_READ, NULL);
	if (hFile == INVALID_HANDLE_VALUE)  return FALSE;
	Offset = GetFileSize(hFile, NULL);
	if (Offset == 0xFFFFFFFF)   // Fehler?
	{
		CloseHandle(hFile);
		return FALSE;
	}
	szBuf = (char*) malloc(Offset+1);
	if (!szBuf)
	{
		CloseHandle(hFile);
		return FALSE;
	}
	if (!ReadFile(hFile, szBuf, Offset, &dwLength, NULL))
	{
		CloseHandle(hFile);
		free(szBuf);
		return FALSE;
	}
	CloseHandle(hFile);
	if (dwLength != Offset)
	{
		free(szBuf);
		return FALSE;
	}
	szBuf[dwLength] = '\0';
	if (strncmp(szBuf, TEXT("Particle Tracer - RGB-Palette:"), 30) != 0)
	{
		free(szBuf);
		return FALSE;
	}
	Offset = 30;
	while ((Offset<dwLength)  &&  (szBuf[Offset]!='\r')  &&  (szBuf[Offset]!='\n'))  Offset++;
	while ((Offset<dwLength)  &&  (szBuf[Offset]=='\r')  ||  (szBuf[Offset]=='\n'))  Offset++;
	if (Offset==dwLength)  return FALSE;
	i = 0;
	while (1)
	{
		iTmp = atoi(szBuf+Offset);
		if (iTmp<0  ||  iTmp>255)
		{
			free(szBuf);
			return FALSE;
		}
		LoadPal[i].rgbRed = iTmp;
		while (szBuf[Offset]>='0'  &&  szBuf[Offset]<='9')
		{
			Offset++;
			if (Offset>=dwLength)
			{
				free(szBuf);
				return FALSE;
			}
		}
		while (szBuf[Offset]==' '  ||  szBuf[Offset]=='\t')
		{
			Offset++;
			if (Offset>=dwLength)
			{
				free(szBuf);
				return FALSE;
			}
		}

		iTmp = atoi(szBuf+Offset);
		if (iTmp<0  ||  iTmp>255)
		{
			free(szBuf);
			return FALSE;
		}
		LoadPal[i].rgbGreen = iTmp;
		while (szBuf[Offset]>='0'  &&  szBuf[Offset]<='9')
		{
			Offset++;
			if (Offset>=dwLength)
			{
				free(szBuf);
				return FALSE;
			}
		}
		while (szBuf[Offset]==' '  ||  szBuf[Offset]=='\t')
		{
			Offset++;
			if (Offset>=dwLength)
			{
				free(szBuf);
				return FALSE;
			}
		}

		iTmp = atoi(szBuf+Offset);
		if (iTmp<0  ||  iTmp>255)
		{
			free(szBuf);
			return FALSE;
		}
		LoadPal[i].rgbBlue = iTmp;
		LoadPal[i].rgbReserved = 0;
		if (i==255)  break;
		while (szBuf[Offset]!='\r'  &&  szBuf[Offset]!='\n')
		{
			Offset++;
			if (Offset>=dwLength)
			{
				free(szBuf);
				return FALSE;
			}
		}
		while (szBuf[Offset]=='\r'  ||  szBuf[Offset]=='\n')
		{
			Offset++;
			if (Offset>=dwLength)
			{
				free(szBuf);
				return FALSE;
			}
		}
		i++;
	}
	free(szBuf);
	return TRUE;
}


BOOL SaveRGBPalette(HBITMAP hBitmap, char *szSaveFile)
{
	RGBQUAD ActPal[256];
	HANDLE  hFile;
	TCHAR   szBuf[1024], szTmp[128];
	DWORD   iResult;
	int     i;

	GetRGBPalette(hBitmap, ActPal);
	hFile = CreateFile(szSaveFile, GENERIC_WRITE, 0, NULL, CREATE_ALWAYS,
		                 FILE_ATTRIBUTE_ARCHIVE | FILE_FLAG_SEQUENTIAL_SCAN, NULL);
	if (hFile == INVALID_HANDLE_VALUE)  return FALSE;
	strcpy(szBuf, TEXT("Particle Tracer - RGB-Palette:\r\n"));
//	strcat(szBuf, TEXT("\tFlche\tMax. GW\tMittl. GW\r\n"));
	if (!WriteFile(hFile, szBuf, strlen(szBuf), &iResult, NULL))
	{
		CloseHandle(hFile);
		return FALSE;
	}
	for (i=0; i<256; i++)
	{
		strcpy(szBuf, ltoa(ActPal[i].rgbRed, szTmp, 10));
		strcat(szBuf, TEXT("\t"));
		strcat(szBuf, ltoa(ActPal[i].rgbGreen, szTmp, 10));
		strcat(szBuf, TEXT("\t"));
		strcat(szBuf, ltoa(ActPal[i].rgbBlue, szTmp, 10));
		strcat(szBuf, TEXT("\r\n"));
		if (!WriteFile(hFile, szBuf, strlen(szBuf), &iResult, NULL))
		{
			CloseHandle(hFile);
			return FALSE;
		}
	}
	CloseHandle(hFile);
	return TRUE;
}


void FlipVertical(HBITMAP hBitmap)
{
	BITMAP bitmap;
	BYTE  *pBits, iTmp;
	long   x, y, yOffset, yOffset2;

	GetObject(hBitmap, sizeof(BITMAP), &bitmap);
	pBits = (BYTE*) bitmap.bmBits;
	for (y=0; y<bitmap.bmHeight/2; y++)
	{ 
		yOffset  = y*bitmap.bmWidthBytes;
		yOffset2 = (bitmap.bmHeight-1-y)*bitmap.bmWidthBytes;
	  for (x=0; x<bitmap.bmWidth; x++)
		{
			iTmp = pBits[yOffset+x];
			pBits[yOffset+x] = pBits[yOffset2+x];
			pBits[yOffset2+x] = iTmp;
		}
	}
}


void FlipHorizontal(HBITMAP hBitmap)
{
	BITMAP bitmap;
	BYTE  *pBits, iTmp;
	long   x, y, yOffset;

	GetObject(hBitmap, sizeof(BITMAP), &bitmap);
	pBits = (BYTE*) bitmap.bmBits;
	for (y=0; y<bitmap.bmHeight; y++)
	{ 
		yOffset = y*bitmap.bmWidthBytes;
		for (x=0; x<bitmap.bmWidth/2; x++)
		{
			iTmp = pBits[yOffset+x];
			pBits[yOffset+x] = pBits[yOffset+bitmap.bmWidth-1-x];
			pBits[yOffset+bitmap.bmWidth-1-x] = iTmp;
		}
	}
}


void Rotate180Degrees(HBITMAP &hBitmap)
{
	BITMAP bitmap;
	BYTE  *pBits, iTmp;
	long   x, y, yOffset, yOffset2;

	GetObject(hBitmap, sizeof(BITMAP), &bitmap);
	pBits = (BYTE*) bitmap.bmBits;
	for (y=0; y<bitmap.bmHeight/2; y++)
	{
		yOffset  = y*bitmap.bmWidthBytes;
		yOffset2 = (bitmap.bmHeight-1-y)*bitmap.bmWidthBytes;
	  for (x=0; x<bitmap.bmWidth; x++)
		{
			iTmp = pBits[yOffset+x];
			pBits[yOffset+x] = pBits[yOffset2+bitmap.bmWidth-1-x];
			pBits[yOffset2+bitmap.bmWidth-1-x] = iTmp;
		}
	}
	if (bitmap.bmHeight & 1)
	{ 
		yOffset = (bitmap.bmHeight/2)*bitmap.bmWidthBytes;
	  for (x=0; x<bitmap.bmWidth/2; x++)
		{
			iTmp = pBits[yOffset+x];
			pBits[yOffset+x] = pBits[y*bitmap.bmWidthBytes+bitmap.bmWidth-1-x];
			pBits[yOffset+bitmap.bmWidth-1-x] = iTmp;
		}
	}
}


BOOL RotateLeft(HBITMAP &hBitmap)
{
	HBITMAP    hNewBitmap;
	BITMAPINFO *pbmi;
	DIBSECTION dibsection;
	RGBQUAD    ActPal[256];
	BYTE       *pBitsSrc, *pBitsDst;
	int        Counter, x, y, yOffset;
	
	GetObject(hBitmap, sizeof(DIBSECTION), &dibsection);
	pBitsSrc = (BYTE*) dibsection.dsBm.bmBits;
	pbmi = (BITMAPINFO*) malloc(dibsection.dsBmih.biSize + 256*sizeof(RGBQUAD));
	if (!pbmi)  return FALSE;
	pbmi->bmiHeader = dibsection.dsBmih;
	pbmi->bmiHeader.biWidth = dibsection.dsBm.bmHeight;
	pbmi->bmiHeader.biHeight= dibsection.dsBm.bmWidth;
	GetRGBPalette(hBitmap, ActPal);
	for (Counter=0; Counter<=255; Counter++)
	{
		pbmi->bmiColors[Counter].rgbBlue = ActPal[Counter].rgbBlue;
		pbmi->bmiColors[Counter].rgbGreen = ActPal[Counter].rgbGreen;
		pbmi->bmiColors[Counter].rgbRed = ActPal[Counter].rgbRed;
		pbmi->bmiColors[Counter].rgbReserved = 0;
	}

	// DIB Section auf Basis dieser Daten anlegen
  hNewBitmap = CreateDIBSection (NULL, pbmi, DIB_RGB_COLORS, (void**) &pBitsDst, NULL, 0) ;

  if (hNewBitmap == NULL)
  {
    free (pbmi) ;
	  return FALSE;
  }

	// Kopieren
	Counter = 4*((dibsection.dsBm.bmHeight*8+31)/32);
	for (y=0; y<dibsection.dsBm.bmHeight; y++)
	{
		yOffset = (dibsection.dsBm.bmHeight-y-1)*dibsection.dsBm.bmWidthBytes;
    for (x=0; x<dibsection.dsBm.bmWidth; x++)
		{
			pBitsDst[x*Counter+y] = pBitsSrc[yOffset+x];
		}
	}
	DeleteObject(hBitmap);
	hBitmap = hNewBitmap;
		
	free(pbmi);
	return TRUE;
}


BOOL RotateRight(HBITMAP &hBitmap)
{
	HBITMAP    hNewBitmap;
	BITMAPINFO *pbmi;
	DIBSECTION dibsection;
	RGBQUAD    ActPal[256];
	BYTE       *pBitsSrc, *pBitsDst;
	int        Counter, x, y, xOffset;
	
	GetObject(hBitmap, sizeof(DIBSECTION), &dibsection);
	pBitsSrc = (BYTE*) dibsection.dsBm.bmBits;
	pbmi = (BITMAPINFO*) malloc(dibsection.dsBmih.biSize + 256*sizeof(RGBQUAD));
	if (!pbmi)  return FALSE;
	pbmi->bmiHeader = dibsection.dsBmih;
	pbmi->bmiHeader.biWidth = dibsection.dsBm.bmHeight;
	pbmi->bmiHeader.biHeight= dibsection.dsBm.bmWidth;
	GetRGBPalette(hBitmap, ActPal);
	for (Counter=0; Counter<=255; Counter++)
	{
		pbmi->bmiColors[Counter].rgbBlue = ActPal[Counter].rgbBlue;
		pbmi->bmiColors[Counter].rgbGreen = ActPal[Counter].rgbGreen;
		pbmi->bmiColors[Counter].rgbRed = ActPal[Counter].rgbRed;
		pbmi->bmiColors[Counter].rgbReserved = 0;
	}

	// DIB Section auf Basis dieser Daten anlegen
  hNewBitmap = CreateDIBSection (NULL, pbmi, DIB_RGB_COLORS, (void**) &pBitsDst, NULL, 0) ;

  if (hNewBitmap == NULL)
  {
    free (pbmi) ;
	  return FALSE;
  }

	// Kopieren
	Counter = 4*((dibsection.dsBm.bmHeight*8+31)/32);
  for (x=0; x<dibsection.dsBm.bmWidth; x++)
	{
		xOffset = (dibsection.dsBm.bmWidth-x-1)*Counter;
		for (y=0; y<dibsection.dsBm.bmHeight; y++)
		{
			pBitsDst[xOffset+y] = pBitsSrc[y*dibsection.dsBm.bmWidthBytes+x];
		}
	}
	DeleteObject(hBitmap);
	hBitmap = hNewBitmap;
		
	free(pbmi);
	return TRUE;
}


BOOL ChangeBitmapSize(HBITMAP &hBitmap, BOOL DeleteOld, long NewWidth, long NewHeight, long OffsetX, long OffsetY, long FillBorderMode)
{
	HBITMAP    hNewBitmap;
	BITMAPINFO *pbmi;
	BITMAP     newbitmap;
	DIBSECTION dibsection;
	RGBQUAD    ActPal[256];
	BYTE       *pBitsSrc, *pBitsDst;
	long       OffsetOld, OffsetNew, x, y, i, j, iDir, jDir;
	double     Avg, AvgLine;
	
	if (FillBorderMode<-4  ||  FillBorderMode>256)  FillBorderMode=0;
	GetObject(hBitmap, sizeof(DIBSECTION), &dibsection);
	pBitsSrc = (BYTE*) dibsection.dsBm.bmBits;
	pbmi = (BITMAPINFO*) malloc(dibsection.dsBmih.biSize + 256*sizeof(RGBQUAD));
	if (!pbmi)  return FALSE;
	pbmi->bmiHeader = dibsection.dsBmih;
	pbmi->bmiHeader.biWidth = NewWidth;
	pbmi->bmiHeader.biHeight= NewHeight;
	GetRGBPalette(hBitmap, ActPal);
	for (x=0; x<=255; x++)
	{
		pbmi->bmiColors[x].rgbBlue  = ActPal[x].rgbBlue;
		pbmi->bmiColors[x].rgbGreen = ActPal[x].rgbGreen;
		pbmi->bmiColors[x].rgbRed   = ActPal[x].rgbRed;
		pbmi->bmiColors[x].rgbReserved = 0;
	}

	// DIB Section auf Basis dieser Daten anlegen
  hNewBitmap = CreateDIBSection (NULL, pbmi, DIB_RGB_COLORS, (void**) &pBitsDst, NULL, 0) ;

  if (hNewBitmap == NULL)
  {
    free (pbmi) ;
	  return FALSE;
  }
	GetObject(hNewBitmap, sizeof(BITMAP), &newbitmap);

	// Kopieren
	for (y=0; y<dibsection.dsBm.bmHeight; y++)
	{
		OffsetOld = y*dibsection.dsBm.bmWidthBytes;
		OffsetNew = (y+OffsetY)*newbitmap.bmWidthBytes+OffsetX;
	  for (x=0; x<dibsection.dsBm.bmWidth; x++)
		{
			pBitsDst[OffsetNew] = pBitsSrc[OffsetOld];
			OffsetOld++;
			OffsetNew++;
		}
	}

	// Rand auffllen
	if (FillBorderMode==256)
	{
		Avg = 0;
		for (y=0; y<dibsection.dsBm.bmHeight; y++)
		{
			OffsetOld = y*dibsection.dsBm.bmWidthBytes;
			AvgLine = 0;
			for (x=0; x<dibsection.dsBm.bmWidth; x++)
			{
				AvgLine += pBitsSrc[OffsetOld];
				OffsetOld++;
			}
			AvgLine /= dibsection.dsBm.bmWidth;
			Avg += AvgLine;
		}
		Avg /= dibsection.dsBm.bmHeight;
		FillBorderMode = long(Avg);
	}
	if (FillBorderMode>=0)
	{
		for (y=0; y<OffsetY; y++)
		{
			OffsetNew = y*newbitmap.bmWidthBytes;
			for (x=0; x<newbitmap.bmWidth; x++)
			{
				pBitsDst[OffsetNew] = (BYTE) FillBorderMode;
				OffsetNew++;
			}
		}
		for (y=OffsetY; y<OffsetY+dibsection.dsBm.bmHeight; y++)
		{
			OffsetNew = y*newbitmap.bmWidthBytes;
			for (x=0; x<OffsetX; x++)
			{
				pBitsDst[OffsetNew] = (BYTE) FillBorderMode;
				OffsetNew++;
			}
			OffsetNew = y*newbitmap.bmWidthBytes+OffsetX+dibsection.dsBm.bmWidth;
			for (x=OffsetX+dibsection.dsBm.bmWidth; x<newbitmap.bmWidth; x++)
			{
				pBitsDst[OffsetNew] = (BYTE) FillBorderMode;
				OffsetNew++;
			}
		}
		for (y=OffsetY+dibsection.dsBm.bmHeight; y<newbitmap.bmHeight; y++)
		{
			OffsetNew = y*newbitmap.bmWidthBytes;
			for (x=0; x<newbitmap.bmWidth; x++)
			{
				pBitsDst[OffsetNew] = (BYTE) FillBorderMode;
				OffsetNew++;
			}
		}
	}

	// Spiegeln inklusive Randpixel
	if (FillBorderMode == -1)
	{
		j    = 0;
		jDir = 1;
		for (y=0; y<OffsetY; y++)
		{
			OffsetNew = (OffsetY-1-y)*newbitmap.bmWidthBytes + OffsetX;
			OffsetOld = j*dibsection.dsBm.bmWidthBytes;
			for (x=0; x<dibsection.dsBm.bmWidth; x++)
			{
				pBitsDst[OffsetNew] = pBitsSrc[OffsetOld];
				OffsetNew++;
				OffsetOld++;
			}
			j += jDir;
			if (j >= dibsection.dsBm.bmHeight)
			{
				j    = dibsection.dsBm.bmHeight - 1;
				jDir = -1;
			}
			if (j<0)
			{
				j    = 0;
				jDir = 1;
			}
		}
		j    = dibsection.dsBm.bmHeight - 1;
		jDir = -1;
		for (y=OffsetY+dibsection.dsBm.bmHeight; y<newbitmap.bmHeight; y++)
		{
			OffsetNew = y*newbitmap.bmWidthBytes + OffsetX;
			OffsetOld = j*dibsection.dsBm.bmWidthBytes;
			for (x=0; x<dibsection.dsBm.bmWidth; x++)
			{
				pBitsDst[OffsetNew] = pBitsSrc[OffsetOld];
				OffsetNew++;
				OffsetOld++;
			}
			j += jDir;
			if (j >= dibsection.dsBm.bmHeight)
			{
				j    = dibsection.dsBm.bmHeight - 1;
				jDir = -1;
			}
			if (j<0)
			{
				j    = 0;
				jDir = 1;
			}
		}
		for (y=0; y<newbitmap.bmHeight; y++)
		{
			OffsetNew = y*newbitmap.bmWidthBytes+OffsetX;
			i    = 0;
			iDir = 1;
			for (x=0; x<OffsetX; x++)
			{
				pBitsDst[OffsetNew-x-1] = pBitsDst[OffsetNew+i];
				i += iDir;
				if (i >= dibsection.dsBm.bmWidth)
				{
					i    = dibsection.dsBm.bmWidth-1;
					iDir = -1;
				}
				if (i<0)
				{
					i    = 0;
					iDir = 1;
				}
			}
			OffsetNew = y*newbitmap.bmWidthBytes+OffsetX;
			i    = dibsection.dsBm.bmWidth - 1;
			iDir = -1;
			for (x=dibsection.dsBm.bmWidth; x<newbitmap.bmWidth-OffsetX; x++)
			{
				pBitsDst[OffsetNew+x] = pBitsDst[OffsetNew+i];
				i += iDir;
				if (i >= dibsection.dsBm.bmWidth)
				{
					i    = dibsection.dsBm.bmWidth-1;
					iDir = -1;
				}
				if (i<0)
				{
					i    = 0;
					iDir = 1;
				}
			}
		}
	}

	// Spiegeln ohne Randpixel
	if (FillBorderMode == -2)
	{
		j    = (dibsection.dsBm.bmHeight>1) ? 1 : 0;
		jDir = (dibsection.dsBm.bmHeight>1) ? 1 : 0;
		for (y=0; y<OffsetY; y++)
		{
			OffsetNew = (OffsetY-1-y)*newbitmap.bmWidthBytes + OffsetX;
			OffsetOld = j*dibsection.dsBm.bmWidthBytes;
			for (x=0; x<dibsection.dsBm.bmWidth; x++)
			{
				pBitsDst[OffsetNew] = pBitsSrc[OffsetOld];
				OffsetNew++;
				OffsetOld++;
			}
			j += jDir;
			if (j >= dibsection.dsBm.bmHeight)
			{
				j    = dibsection.dsBm.bmHeight - 2;
				jDir = -1;
			}
			if (j<0)
			{
				j    = 1;
				jDir = 1;
			}
		}
		j    = (dibsection.dsBm.bmHeight>1) ? dibsection.dsBm.bmHeight - 2 : 0;
		jDir = (dibsection.dsBm.bmHeight>1) ? -1 : 0;
		for (y=OffsetY+dibsection.dsBm.bmHeight; y<newbitmap.bmHeight; y++)
		{
			OffsetNew = y*newbitmap.bmWidthBytes + OffsetX;
			OffsetOld = j*dibsection.dsBm.bmWidthBytes;
			for (x=0; x<dibsection.dsBm.bmWidth; x++)
			{
				pBitsDst[OffsetNew] = pBitsSrc[OffsetOld];
				OffsetNew++;
				OffsetOld++;
			}
			j += jDir;
			if (j >= dibsection.dsBm.bmHeight)
			{
				j    = dibsection.dsBm.bmHeight - 2;
				jDir = -1;
			}
			if (j<0)
			{
				j    = 1;
				jDir = 1;
			}
		}
		for (y=0; y<newbitmap.bmHeight; y++)
		{
			OffsetNew = y*newbitmap.bmWidthBytes+OffsetX;
			i    = (dibsection.dsBm.bmWidth>1) ? 1 : 0;
  		iDir = (dibsection.dsBm.bmWidth>1) ? 1 : 0;
			for (x=0; x<OffsetX; x++)
			{
				pBitsDst[OffsetNew-x-1] = pBitsDst[OffsetNew+i];
				i += iDir;
				if (i >= dibsection.dsBm.bmWidth)
				{
					i    = dibsection.dsBm.bmWidth - 2;
					iDir = -1;
				}
				if (i<0)
				{
					i    = 1;
					iDir = 1;
				}
			}
			OffsetNew = y*newbitmap.bmWidthBytes+OffsetX;
			i    = (dibsection.dsBm.bmWidth>1) ? dibsection.dsBm.bmWidth - 2 : 0;
  		iDir = (dibsection.dsBm.bmWidth>1) ? -1 : 0;
			for (x=dibsection.dsBm.bmWidth; x<newbitmap.bmWidth-OffsetX; x++)
			{
				pBitsDst[OffsetNew+x] = pBitsDst[OffsetNew+i];
				i += iDir;
				if (i >= dibsection.dsBm.bmWidth)
				{
					i    = dibsection.dsBm.bmWidth - 2;
					iDir = -1;
				}
				if (i<0)
				{
					i    = 1;
					iDir = 1;
				}
			}
		}
	}

	// Periodisch wiederholen
	if (FillBorderMode == -3)
	{
		j = dibsection.dsBm.bmHeight - 1;
		for (y=0; y<OffsetY; y++)
		{
			OffsetNew = (OffsetY-1-y)*newbitmap.bmWidthBytes + OffsetX;
			OffsetOld = j*dibsection.dsBm.bmWidthBytes;
			for (x=0; x<dibsection.dsBm.bmWidth; x++)
			{
				pBitsDst[OffsetNew] = pBitsSrc[OffsetOld];
				OffsetNew++;
				OffsetOld++;
			}
			j--;
			if (j<0)	j = dibsection.dsBm.bmHeight - 1;
		}
		j = 0;
		for (y=OffsetY+dibsection.dsBm.bmHeight; y<newbitmap.bmHeight; y++)
		{
			OffsetNew = y*newbitmap.bmWidthBytes + OffsetX;
			OffsetOld = j*dibsection.dsBm.bmWidthBytes;
			for (x=0; x<dibsection.dsBm.bmWidth; x++)
			{
				pBitsDst[OffsetNew] = pBitsSrc[OffsetOld];
				OffsetNew++;
				OffsetOld++;
			}
			j++;
			if (j >= dibsection.dsBm.bmHeight)  j = 0;
		}
		for (y=0; y<newbitmap.bmHeight; y++)
		{
			OffsetNew = y*newbitmap.bmWidthBytes+OffsetX;
			i = dibsection.dsBm.bmWidth - 1;
			for (x=0; x<OffsetX; x++)
			{
				pBitsDst[OffsetNew-x-1] = pBitsDst[OffsetNew+i];
				i--;
				if (i<0)	i = dibsection.dsBm.bmWidth;
			}
			OffsetNew = y*newbitmap.bmWidthBytes+OffsetX;
			i = 0;
			for (x=dibsection.dsBm.bmWidth; x<newbitmap.bmWidth-OffsetX; x++)
			{
				pBitsDst[OffsetNew+x] = pBitsDst[OffsetNew+i];
				i++;
				if (i >= dibsection.dsBm.bmWidth)  i = 0;
			}
		}
	}

	// Lokale Mittelwerte
	if (FillBorderMode == -4)
	{
		long x1=OffsetX, x2=OffsetX+dibsection.dsBm.bmWidth -1;
		long y1=OffsetY, y2=OffsetY+dibsection.dsBm.bmHeight-1;

		while (x1>0  ||  y1>0  ||  x2<newbitmap.bmWidth-1  ||  y2<newbitmap.bmHeight-1)
		{
			if (x1>0)
			{
				OffsetNew = (y1+1)*newbitmap.bmWidthBytes+x1;
				for (y=0; y<y2-y1-1; y++)
				{
					Avg  = pBitsDst[OffsetNew-newbitmap.bmWidthBytes]+pBitsDst[OffsetNew]+pBitsDst[OffsetNew+newbitmap.bmWidthBytes];
					Avg /= 3.0;
					pBitsDst[OffsetNew-1] = (BYTE) Avg;
					OffsetNew += newbitmap.bmWidthBytes;
				}
				OffsetNew = y1*newbitmap.bmWidthBytes+x1;
				Avg  = pBitsDst[OffsetNew];
				if (y1<y2)
				{
					Avg += pBitsDst[OffsetNew+newbitmap.bmWidthBytes];
				  Avg /= 2.0;
				}
				pBitsDst[OffsetNew-1] = (BYTE) Avg;
				OffsetNew = y2*newbitmap.bmWidthBytes+x1;
				Avg = pBitsDst[OffsetNew];
				if (y1<y2)
				{
					Avg += pBitsDst[OffsetNew-newbitmap.bmWidthBytes];
					Avg /= 2.0;
				}
				pBitsDst[OffsetNew-1] = (BYTE) Avg;
			}
			if (x2<newbitmap.bmWidth-1)
			{
				OffsetNew = (y1+1)*newbitmap.bmWidthBytes+x2;
				for (y=0; y<y2-y1-1; y++)
				{
					Avg = pBitsDst[OffsetNew-newbitmap.bmWidthBytes]+pBitsDst[OffsetNew]+pBitsDst[OffsetNew+newbitmap.bmWidthBytes];
					Avg /= 3.0;
					pBitsDst[OffsetNew+1] = (BYTE) Avg;
					OffsetNew += newbitmap.bmWidthBytes;
				}
				OffsetNew = y1*newbitmap.bmWidthBytes+x2;
				Avg  = pBitsDst[OffsetNew];
				if (y1<y2)
				{
					Avg += pBitsDst[OffsetNew+newbitmap.bmWidthBytes];
					Avg /= 2.0;
				}
				pBitsDst[OffsetNew+1] = (BYTE) Avg;
				OffsetNew = y2*newbitmap.bmWidthBytes+x2;
				Avg  = pBitsDst[OffsetNew];
				if (y1<y2)
				{
					Avg +=pBitsDst[OffsetNew-newbitmap.bmWidthBytes];
					Avg /= 2.0;
				}
				pBitsDst[OffsetNew+1] = (BYTE) Avg;
			}
			if (y1>0)
			{
				OffsetNew = y1*newbitmap.bmWidthBytes+x1+1;
				for (x=0; x<x2-x1-1; x++)
				{
					Avg = pBitsDst[OffsetNew-1]+pBitsDst[OffsetNew]+pBitsDst[OffsetNew+1];
					Avg /= 3.0;
					pBitsDst[OffsetNew-newbitmap.bmWidthBytes] = (BYTE) Avg;
					OffsetNew++;
				}
				OffsetNew = y1*newbitmap.bmWidthBytes+x1;
				Avg  = pBitsDst[OffsetNew];
				if (x1<x2)
				{
					Avg += pBitsDst[OffsetNew+1];
					Avg /= 2.0;
				}
				pBitsDst[OffsetNew-newbitmap.bmWidthBytes] = (BYTE) Avg;
				OffsetNew = y1*newbitmap.bmWidthBytes+x2;
				Avg  = pBitsDst[OffsetNew];
				if (x1<x2)
				{
					Avg += pBitsDst[OffsetNew-1];
					Avg /= 2.0;
				}
				pBitsDst[OffsetNew-newbitmap.bmWidthBytes] = (BYTE) Avg;
			}
			if (y2<newbitmap.bmWidth)
			{
				OffsetNew = y2*newbitmap.bmWidthBytes+x1+1;
				for (x=0; x<x2-x1-1; x++)
				{
					Avg = pBitsDst[OffsetNew-1]+pBitsDst[OffsetNew]+pBitsDst[OffsetNew+1];
					Avg /= 3.0;
					pBitsDst[OffsetNew+newbitmap.bmWidthBytes] = (BYTE) Avg;
					OffsetNew++;
				}
				OffsetNew = y2*newbitmap.bmWidthBytes+x1;
				Avg  = pBitsDst[OffsetNew];
				if (x1<x2)
				{
					Avg += pBitsDst[OffsetNew+1];
					Avg /= 2.0;
				}
				pBitsDst[OffsetNew+newbitmap.bmWidthBytes] = (BYTE) Avg;
				OffsetNew = y2*newbitmap.bmWidthBytes+x2;
				Avg  = pBitsDst[OffsetNew];
				if (x1<x2)
				{
					Avg += pBitsDst[OffsetNew-1];
					Avg /= 2.0;
				}
				pBitsDst[OffsetNew+newbitmap.bmWidthBytes] = (BYTE) Avg;
			}
			if (x1>0  &&  y1>0)
				pBitsDst[(y1-1)*newbitmap.bmWidthBytes+x1-1] = 
				  (pBitsDst[(y1-1)*newbitmap.bmWidthBytes+x1]  +  pBitsDst[y1*newbitmap.bmWidthBytes+x1-1])/2;
			if (x2<newbitmap.bmWidth-1  &&  y1>0)
				pBitsDst[(y1-1)*newbitmap.bmWidthBytes+x2+1] = 
				  (pBitsDst[(y1-1)*newbitmap.bmWidthBytes+x2]  +  pBitsDst[y1*newbitmap.bmWidthBytes+x2+1])/2;
			if (x1>0  &&  y2<newbitmap.bmHeight-1)
				pBitsDst[(y2+1)*newbitmap.bmWidthBytes+x1-1] = 
				  (pBitsDst[(y2+1)*newbitmap.bmWidthBytes+x1]  +  pBitsDst[y2*newbitmap.bmWidthBytes+x1-1])/2;
			if (x2<newbitmap.bmWidth-1  &&  y2<newbitmap.bmHeight-1)
				pBitsDst[(y2+1)*newbitmap.bmWidthBytes+x2+1] = 
				  (pBitsDst[(y2+1)*newbitmap.bmWidthBytes+x2]  +  pBitsDst[y2*newbitmap.bmWidthBytes+x2+1])/2;

			if (x1>0) x1--;
			if (x2<newbitmap.bmWidth -1) x2++;
			if (y1>0) y1--;
			if (y2<newbitmap.bmHeight-1) y2++;
		}
	}

	// Fertig
	if (DeleteOld)  DeleteObject(hBitmap);
	hBitmap = hNewBitmap;
	free(pbmi);
	return TRUE;
}



void SetBMPOrigin(HBITMAP hBitmap, long x, long y)
{
	BITMAP bitmap;
	BYTE   *pBits, *pMemBits;
	long   j, bh, th, lw, rw, OffsetS, OffsetD;

	if (x==0  &&  y==0)  return;
	if (!hBitmap)  return;
	GetObject(hBitmap, sizeof(BITMAP), &bitmap);
	pBits = (BYTE*) bitmap.bmBits;
	pMemBits = (BYTE*) malloc(bitmap.bmWidthBytes*bitmap.bmHeight);
	if (!pMemBits)  return;
  CopyMemory(pMemBits, pBits, bitmap.bmWidthBytes*bitmap.bmHeight);
	
	bh = y;
	th = bitmap.bmHeight-y;
	lw = x;
	rw = bitmap.bmWidth-x;
	OffsetS = y*bitmap.bmWidthBytes;
	OffsetD = 0;
	for (j=0; j<th; j++)
	{
		CopyMemory(pBits+OffsetD, pMemBits+OffsetS+lw, rw);
		CopyMemory(pBits+OffsetD+rw, pMemBits+OffsetS, lw);
		OffsetS += bitmap.bmWidthBytes;
		OffsetD += bitmap.bmWidthBytes;
	}
	OffsetS = 0;
	OffsetD = th*bitmap.bmWidthBytes;
	for (j=0; j<bh; j++)
	{
		CopyMemory(pBits+OffsetD, pMemBits+OffsetS+lw, rw);
		CopyMemory(pBits+OffsetD+rw, pMemBits+OffsetS, lw);
		OffsetS += bitmap.bmWidthBytes;
		OffsetD += bitmap.bmWidthBytes;
	}
	free (pMemBits);
}